(* $Id: pax.ml 1085 2012-07-22 22:36:09Z gerd $ 
 * ----------------------------------------------------------------------
 * Copyright 2008 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Pax_types
open Printf

type in_archive =
    { in_ar : archive;
      in_next : unit -> archive_rd option;
      in_close : unit -> unit;
      in_postprocess : (Godi_file.filename, Godi_file.filename->unit) Hashtbl.t
    }

type in_element =
 in_archive * archive_rd


type in_file =
    [ `File of Godi_file.filename
    | `Stdin
    | `Obj_channel of Netchannels.in_obj_channel
    ]


type compression_kind =
    [ `None
    | `Gzip
    | `Bzip2
    ]

type format =
    [ `Tar
    ]

type overwrite_mode =
    [ `Always
    | `Update
    | `Ignore
    | `Reject
    ]


(* Win32: The gzip/bzip2 commands are started using command.com, so we have
   to:
    - translate the command paths to native format
    - use the system-specific Filename.quote for quoting names
 *)

let gzip_cmd = lazy (
  let path =
    if !Godi_boot.is_bootconsole then
      "gzip"
    else
      Godi_conf.cmd_gzip in
  Godi_file.filename path
)



let bzip2_cmd = lazy (
  let path =
    Godi_conf.cmd_bzip2 in
  Godi_file.filename path
)


let mk_gzip_cmd filename_opt  =
  let cmd_path = Godi_file.nativename (Lazy.force gzip_cmd) in
  sprintf "%s -c >%s"
    cmd_path
    (match filename_opt with
       | None -> ""
       | Some fn -> 
	   let name' = Filename.quote fn in
	   if fn <> "" && fn.[0] = '-' then
	     "-- "  ^ name'
	   else
	     name'
    )



let mk_gunzip_cmd filename_opt =
  let cmd_path = Godi_file.nativename (Lazy.force gzip_cmd) in
  sprintf "%s -cd %s"
    cmd_path
    (match filename_opt with
       | None -> ""
       | Some fn -> 
	   let name' = Filename.quote fn in
	   if fn <> "" && fn.[0] = '-' then
	     "-- "  ^ name'
	   else
	     name'
    )


let mk_bzip2_cmd filename_opt  =
  let cmd_path = Godi_file.nativename (Lazy.force bzip2_cmd) in
  sprintf "%s -c >%s"
    cmd_path
    (match filename_opt with
       | None -> ""
       | Some fn -> 
	   let name' = Filename.quote fn in
	   if fn <> "" && fn.[0] = '-' then
	     "-- "  ^ name'
	   else
	     name'
    )



let mk_bunzip2_cmd filename_opt =
  let cmd_path = Godi_file.nativename (Lazy.force bzip2_cmd) in
  sprintf "%s -cd %s"
    cmd_path
    (match filename_opt with
       | None -> ""
       | Some fn -> 
	   let name' = Filename.quote fn in
	   if fn <> "" && fn.[0] = '-' then
	     "-- "  ^ name'
	   else
	     name'
    )


let string_of_exit_code =
  function
    | Unix.WEXITED n -> "WEXITED " ^ string_of_int n
    | Unix.WSIGNALED n -> "WSIGNALED " ^ string_of_int n
    | Unix.WSTOPPED n -> "WSTOPPED " ^ string_of_int n


class input_command cmd =
  let ch = Unix.open_process_in cmd in
object (self)
  inherit Netchannels.input_channel ch as super

  val mutable closed = false

  method close_in() =
    if not closed then (
      Pax_tar.skip (self :> Netchannels.in_obj_channel);
        (* skip over all remaining data *)
      let p = Unix.close_process_in ch in
      closed <- true;
      if p <> Unix.WEXITED 0 then
        let msg =
          sprintf "Command returns abnormally (%s): %s"
            (string_of_exit_code p)
            cmd in
        let ign =
          try Sys.getenv "GODI_IGNORE_GZIP_ERROR" = "true"
          with _ -> false in
        if ign then (
          prerr_endline (msg ^ " (ignored)")
        )
        else
          raise (Sys_error msg)
    )
end
;;


class output_command cmd =
  let ch = Unix.open_process_out cmd in
object (self)
  inherit Netchannels.output_channel ch as super

  val mutable closed = false

  method close_out() =
    if not closed then (
      let p = Unix.close_process_out ch in
      closed <- true;
      if p <> Unix.WEXITED 0 then
        raise (Sys_error(
                 sprintf "Command returns abnormally (%s): %s"
                   (string_of_exit_code p)
                   cmd))
    )
end
;;


let copy_channel 
      len 
      (src_ch : Netchannels.in_obj_channel) 
      (dest_ch : Netchannels.out_obj_channel) =
  (* Copies contents from src_ch to dest_ch. Returns [true] if at EOF.
   *)
  let buf = String.create 4096 in
  let slen = String.length buf in
  let k = ref 0L in
  try
    while true do
      let m = Int64.to_int (min (Int64.of_int slen) (Int64.sub len !k)) in
      if m = 0 then raise Exit;
      let n = src_ch # input buf 0 m in
      if n = 0 then raise Sys_blocked_io;
      dest_ch # really_output buf 0 n;
      k := Int64.add !k (Int64.of_int n)
    done;
    assert false
  with
    | End_of_file
    | Exit ->
	()


let read_gzip =
  ref (fun name_opt ->
         new input_command (mk_gunzip_cmd name_opt)
      )


let read in_file compression fmt =
  let (ch, close) =
    match (in_file, compression) with
      | (`File name), `None ->
	  let ch = 
	    new Netchannels.input_channel (Godi_file.open_in `Binary name) in
	  (ch, ch#close_in)
      | `Stdin, `None ->
	  set_binary_mode_in stdin true;
	  let ch = 
	    new Netchannels.input_channel stdin in
	  (ch, (fun () -> Pax_tar.skip ch))
      | (`Obj_channel ch), `None ->
	  (ch, ch#close_in)
      | (`File name), `Gzip ->
	  let ch = !read_gzip (Some (Godi_file.nativename name)) in
	  (ch, (fun () -> ch#close_in()))
      | `Stdin,`Gzip ->
	  let ch = !read_gzip None in
	  (ch, (fun () -> ch#close_in()))
      | (`File name), `Bzip2 ->
	  let ch = 
	    new input_command 
              (mk_bunzip2_cmd (Some (Godi_file.nativename name))) in
	  (ch, (fun () -> ch#close_in()))
      | `Stdin,`Bzip2 ->
	  let ch = new input_command (mk_bunzip2_cmd None) in
	  (ch, (fun () -> ch#close_in()))
      | (`Obj_channel _), (`Gzip | `Bzip2) ->
	  failwith 
	    "Pax.read: object channels are incompatible with compression"
  in
  let ar_file =
    ( object
	method read_file() =
	  ch
	method create_file() =
	  assert false
      end
    ) in
  let ar =
    match fmt with
      | `Tar ->
	  new Pax_tar.tar ar_file in
  let cur_el_opt = ref (lazy(ar # read_archive())) in
  { in_ar = ar;
    in_next = (fun () ->
		 match Lazy.force !cur_el_opt with
		   | None -> None
		   | Some el -> 
		       cur_el_opt := lazy (el # next());
		       Some el
	      );
    in_postprocess = Hashtbl.create 50;
    in_close = close
  }


let read_element in_ar =
  match in_ar.in_next() with
    | None -> None
    | Some in_el -> Some (in_ar, in_el)


let get_metadata (_,in_el) =
  in_el # get_metadata

let extract_element_to_memory (_, in_el) =
  let buf = Buffer.create 1000 in
  let buf_ch = new Netchannels.output_buffer buf in
  ( match in_el # get_data() with
      | `Channel data_ch ->
	  buf_ch # output_channel data_ch;
	  Buffer.contents buf
      | `File(file, _) ->
	  Godi_file.read_file ~mode:`Binary file
  )

let slash_re = Str.regexp "/+"

let is_bad_pathname name =
  let l = Str.split_delim slash_re name in
  (l <> [] && List.hd l = "") ||
    (List.mem ".." l) ||
    ( match Sys.os_type with
	| "Cygwin" | "Win32" ->
	    (* Additionally refuse:
                - names containing backslashes (would be interpreted as
                  separator)
                - names containing colons (confusion with names like
                  c:file)
	     *)
	    List.exists 
	      (fun p -> String.contains p '\\' || String.contains p ':')
	      l
	| _ ->
	    false
    )


let iter_dirs f1 f2 split_name =
  (* Invoke f1 on all inner non-root paths, in top-to-bottom order, and f2 on
     the leaf name
   *)
  let rec iter constr acc name_list =
    match name_list with
      | [] ->
	  ()
      | [ last_name ]
      | [ last_name; "" ] ->
	  f2 (constr (List.rev (last_name :: acc)))
      | inner_name :: name_list' ->
	  f1 (constr (List.rev (inner_name :: acc)));
	  iter constr (inner_name :: acc) name_list'
  in
  match split_name with
    | `Relative name_list ->
	iter 
	  (fun l -> Godi_file.combine_filename (`Relative l)) 
	  [] name_list
    | `POSIX_absolute name_list ->
	iter 
	  (fun l -> Godi_file.combine_filename (`POSIX_absolute l)) 
	  [] name_list
    | `WIN32_drive(letter, name_list) ->
	iter 
	  (fun l -> Godi_file.combine_filename (`WIN32_drive(letter,l))) 
	  [] name_list
    | `WIN32_remote(host, name_list) ->
	iter 
	  (fun l -> Godi_file.combine_filename (`WIN32_remote(host,l))) 
	  [] name_list
    | `Error ->
	failwith "Pax.iter_dirs"


let create_dir dirpath =
  if not(Godi_file.file_is_dir dirpath) then (
    if Godi_file.file_exists dirpath then
      failwith ("Not a directory: " ^ Godi_file.slashname dirpath);
    Godi_file.create_dir dirpath
  )

let set_mtime path mtime =
  try
    if Sys.os_type = "Win32" then (
      (* - does not work for directories
       * - needs write privilege for file
       *)
      let st = Godi_file.stat ~symlink:true path in
      if st.Unix.LargeFile.st_kind = Unix.S_REG then (
	Godi_file.set_mode path 0o777;
	Godi_file.set_mtime path (Int64.to_float mtime);
	Godi_file.set_mode path st.Unix.LargeFile.st_perm;
      )
    )
    else
      Godi_file.set_mtime path (Int64.to_float mtime)
  with
    | e ->
	prerr_endline("Failed to set mtime for: " ^ 
			Godi_file.slashname path);
	prerr_endline("  (Error: " ^ Printexc.to_string e ^ ")")


let norm_mode_bits is_dir bits =
  (* We look only at the user bits, and support only the combinations:
     r--: read-only, non-executable  (not for directories)
     rw-: read-write, non-executable (not for directories)
     r-x: read-only, executable      (not for directories)
     rwx: read-write, executable

     The bits for group and worls are set to this, too, and it is expected
     that the umask deletes bits.
   *)
  let is_writable = is_dir || (bits land 0o200) <> 0 in
  let is_executable = is_dir || (bits land 0o100) <> 0 in
  let b = 
    4 lor
      (if is_writable then 2 else 0) lor
      (if is_executable then 1 else 0) in
  (b lsl 6) lor (b lsl 3) lor b


let extract_element_as_file ?(insecure = false)
                            ?(prefer_link = false)
                            ?prefix 
                            ?(subst = fun s -> s) 
                            ?(preserve_mtime = false)
                            ?(overwrite = `Always)
                            (in_ar, (in_el : archive_rd)) =
  let m = in_el # get_metadata in
  let ok =
    match m # filetype with
      | `Regular -> true
      | `Directory -> true
      | `Hardlink _ -> true
      | `Symlink _ -> true
      | _ -> false in
  if ok then (
    if not insecure && is_bad_pathname m#name then
      failwith ("Refusing to extract this path name: " ^ m#name);
    let mapped_name = subst m#name in
    if mapped_name <> "" then (
      let local_name =
	match prefix with
	  | None -> 
	      Godi_file.filename mapped_name
	  | Some p -> 
	      Godi_file.concat p (Godi_file.filename mapped_name) in
      let split_name = 
	Godi_file.split_filename local_name in
      iter_dirs
	create_dir
	(fun filepath ->
	   match m # filetype with
	     | `Directory ->
		 create_dir filepath;
		 
		 if preserve_mtime then (
		   Hashtbl.replace 
		     in_ar.in_postprocess
		     filepath
		     (fun p -> set_mtime p m#mtime)
		 )
		     
	     | `Regular
	     | `Hardlink _
	     | `Symlink _ ->

		 let data_pnt = in_el#get_data() in
		 let close_data =
		   match data_pnt with
		     | `File(_,_) -> ref (fun () -> ())
		     | `Channel ch -> ref (fun () -> ch#close_in()) in

		 let is_self_copy =
		   ( match data_pnt with
		       | `File (_, src_file_id) ->
			   if Godi_file.file_exists filepath then
			     let dest_file_id = 
			       Godi_file.file_id ~symlink:true filepath in
			     src_file_id = dest_file_id
			       (* Symlink: It is ok when we replace a symlink
                                  by a copy of the linked file
				*)
			   else
			     false 
		       | `Channel _ ->
			   false
		   ) in

		 let do_overwrite, do_unlink =
		   if is_self_copy then (
		     prerr_endline ("Refusing to overwrite file by itself: " ^ 
				      Godi_file.slashname filepath);
		     (false, false)
		   )
		   else
		     if Godi_file.file_exists ~symlink:true filepath then
		       (* Symlink: a file also exists if it is a symlink to a
                          non-existing file
			*)
		       match overwrite with
			 | `Always -> 
			     (true,true)
			 | `Update ->
			     let t = 
			       Godi_file.file_mtime ~symlink:true filepath in
			     if  t < Int64.to_float m#mtime then
			       (true,true)
			     else
			       (false,false)
			 | `Ignore -> 
			     (false,false) 
			 | `Reject -> 
			     failwith("File exists: " ^ 
					Godi_file.slashname filepath)
		     else
		       (true, false) in
		 if do_overwrite then (
		   if do_unlink then
		     Godi_file.remove filepath;
		   
		   ( match m # filetype with
		       | `Regular ->
			   let do_copy = ref true in
			   if prefer_link then (
			     match data_pnt with
			       | `File(srcname,_) ->
				   ( try
				       Godi_file.create_link srcname filepath;
				       do_copy := false
				     with _ -> ()
				   )
			       | `Channel _ -> ()
			   );
			   if !do_copy then (
			     let in_ch =
			       match data_pnt with
				 | `File(srcname,_) ->
				     let ich =
				       new Netchannels.input_channel
					 (Godi_file.open_in `Binary srcname) in
				     close_data := (fun () -> ich#close_in());
				     ich
				 | `Channel ch -> ch in
			     let filemode = norm_mode_bits false m#mode in
			     let ch = 
			       Godi_file.open_out_gen
				 [Open_wronly; Open_creat; Open_binary]
				 filemode
				 filepath in
			     let out_ch = new Netchannels.output_channel ch in
			     copy_channel m#size in_ch out_ch;
			     out_ch # close_out();
			   
			     if preserve_mtime then
			       set_mtime filepath m#mtime
			   )
			       
		       | `Hardlink link ->
			   Godi_file.create_link
			     (Godi_file.filename link) filepath
			     
		       | `Symlink link ->
			   Godi_file.create_symlink 
			     (Godi_file.filename link) filepath

		       | _ ->
			   ()
		   );
		 );
		 !close_data()
	     | _ -> 
		 assert false
	)
	split_name
    )

  )


let postprocess_ht ht =
  Hashtbl.iter
    (fun path action ->
       action path
    )
    ht;
  Hashtbl.clear ht


let postprocess_extraction in_ar =
  postprocess_ht in_ar.in_postprocess


let close_in in_ar =
  in_ar.in_close()


type out_archive = 
    { out_ar : archive;
      out_wr : archive_wr;
      out_close : unit -> unit
    }

type out_file =
    [ `File of Godi_file.filename
    | `Stdout
    | `Obj_channel of Netchannels.out_obj_channel
    ]


let write_gzip =
  ref (fun name_opt ->
         new output_command (mk_gzip_cmd name_opt)
      )


let create out_file compression fmt =
  let (ch, close) =
    match (out_file, compression) with
      | (`File name), `None ->
	  let ch =
	    new Netchannels.output_channel (Godi_file.open_out `Binary name) in
	  (ch, ch#close_out)
      | `Stdout, `None ->
	  set_binary_mode_out stdout true;
	  let ch = 
	    new Netchannels.output_channel stdout in
	  (ch, (fun () -> ()))
      | (`Obj_channel ch), `None ->
	  (ch, ch#close_out)
      | (`File name), `Gzip ->
	  let ch = !write_gzip (Some (Godi_file.nativename name)) in
	  (ch, ch#close_out)
      | `Stdout,`Gzip ->
	  let ch = !write_gzip None in
	  (ch, ch#close_out)
      | (`File name), `Bzip2 ->
	  let ch = 
	    new output_command 
              (mk_bzip2_cmd (Some (Godi_file.nativename name))) in
	  (ch, ch#close_out)
      | `Stdout,`Bzip2 ->
	  let ch = new output_command (mk_bzip2_cmd None) in
	  (ch, ch#close_out)
      | (`Obj_channel _), (`Gzip | `Bzip2) ->
	  failwith 
	    "Pax.create: object channels are incompatible with compression"
  in
  let ar_file =
    ( object
	method read_file() =
	  assert false
	method create_file() =
	  ch
      end
    ) in
  let ar =
    match fmt with
      | `Tar ->
	  new Pax_tar.tar ar_file in
  let ar_wr = ar # create_archive() in
  { out_ar = ar;
    out_wr = ar_wr;
    out_close = close
  }

let close_out out_ar =
  out_ar.out_wr # close();
  out_ar.out_close()

let add_element_from_file ?prefix ?(subst=fun s -> s) out_ar filename =
  let subst_name = subst filename in
  if subst_name <> "" then (
    let local_name =
      match prefix with
	| None -> 
	    Godi_file.filename filename
	| Some p -> 
	    Godi_file.concat p (Godi_file.filename filename) in
    let st = Godi_file.stat ~symlink:true local_name in
    let filetype_opt =
      match st.Unix.LargeFile.st_kind with
	| Unix.S_REG ->
	    Some `Regular
	| Unix.S_DIR ->
	    Some `Directory
	| Unix.S_LNK ->
	    Some (`Symlink (Godi_file.read_symlink local_name))
	| _ ->
	    None in
    match filetype_opt with
      | None ->
	  ()   (* Ignore these *)
      | Some filetype ->
	  let m =
	    ( object 
		method name = subst_name
		method mode = st.Unix.LargeFile.st_perm
		method size = st.Unix.LargeFile.st_size
		method owner_uid = None
		method owner_uname = None
		method owner_gid = None
		method owner_gname = None
		method mtime = Int64.of_float st.Unix.LargeFile.st_mtime
		method filetype = filetype
		method format = None
		method have_data = false  (* ignored anyway *)
	      end : metadata
	    ) in
	  if filetype = `Regular && out_ar.out_ar#archive_type <> "clipboard" 
	  then (
	    let data_ch = 
	      out_ar.out_wr # put_data m in
	    let file_ch = 
	      new Netchannels.input_channel
		(Godi_file.open_in `Binary local_name) in
	    copy_channel m#size file_ch data_ch;
	    file_ch # close_in();
	    data_ch # close_out();
	  )
	  else
	    out_ar.out_wr # put_metadata_only m
  )


let add_element_from_string ~name ~mode ~mtime ~filetype out_ar data =
  let m =
    ( object 
	method name = name
	method mode = mode
	method size = Int64.of_int (String.length data)
	method owner_uid = None
	method owner_uname = None
	method owner_gid = None
	method owner_gname = None
	method mtime = mtime
	method filetype = filetype
	method format = None
	method have_data = false  (* ignored anyway *)
      end : metadata
    ) in
  let data_ch = out_ar.out_wr # put_data m in
  data_ch # output_string data;
  data_ch # close_out()

  
type file_iterator = unit -> string option


let check_non_win32 fn =
  match Godi_file.split_filename fn with
    | `WIN32_drive(_,_)
    | `WIN32_remote(_,_) ->
	failwith ("WIN32 names not supported in iterations: " ^ 
		    Godi_file.slashname fn)
    | _ ->
	()

let expand_directories ?prefix (f : file_iterator) =

  let current_fetcher = ref (fun () -> assert false) in

  let rec fetch_from_dir localname filename followup =
    let files = ref (Godi_file.read_dir localname) in
    fun () ->
      match !files with
	| file :: files' ->
	    files := files';
	    let fname = filename ^ "/" ^ file in
	    let lname = Godi_file.addname localname file in
	    if Godi_file.file_type ~is_dir:true ~is_not_symlink:true lname
	    then (
	      let old_fetcher = !current_fetcher in
	      current_fetcher := fetch_from_dir lname fname old_fetcher;
	      Some fname
	    )
	    else
	      Some fname
	| [] ->
	    current_fetcher := followup;
	    followup()
  in

  let fetch_from f followup () =
    let filename_opt = f() in
    match filename_opt with
      | None ->
	  current_fetcher := followup;
	  followup()
      | Some filename ->
	  let parsed_filename = Godi_file.filename filename in
	  check_non_win32 parsed_filename;
	  let localname =
	    match prefix with
	      | None -> parsed_filename
	      | Some p -> Godi_file.concat p parsed_filename in
	  if Godi_file.file_type ~is_dir:true ~is_not_symlink:true localname
	  then (
	    let old_fetcher = !current_fetcher in
	    current_fetcher := fetch_from_dir localname filename old_fetcher;
	    Some filename
	  )
	  else
	    Some filename
  in

  let return_end() =
    None
  in

  current_fetcher := fetch_from f return_end;

  (fun () ->
     !current_fetcher()
  )


let file_iterator_of_list ?(nodescent=false) ?prefix list =
  let l = ref list in
  let f () =
    match !l with
      | name :: l' ->
	  l := l';
	  Some name
      | [] ->
	  None in
  if nodescent then
    f
  else
    expand_directories ?prefix f

	  
let file_iterator_of_channel ?(nodescent=false) ?prefix ch =
  let f() =
    try
      Some(ch # input_line())
    with
      | End_of_file ->
	  None
  in
  if nodescent then
    f
  else
    expand_directories ?prefix f


let next_file f =
  f()


type copier =
    { cp_postprocess : (Godi_file.filename, Godi_file.filename->unit) Hashtbl.t
    }

let create_copier() =
  { cp_postprocess = Hashtbl.create 50 }

let copy_file ?src_prefix ?(prefer_link=false)
              ?(subst=fun s -> s) ?(preserve_mtime = false)
              ?(overwrite = `Always)
              cp src dir =

  if not (Godi_file.file_is_dir dir) then
    raise(Sys_error ("Not a directory: " ^ Godi_file.slashname dir));

  let ar = new Pax_clipboard.clipboard src_prefix in
  let out_ar =
    { out_ar = ar;
      out_wr = ar # create_archive();
      out_close = fun() ->()
    } in
  add_element_from_file ?prefix:src_prefix ~subst out_ar src;
  let in_el_opt = ar # read_archive() in
  match in_el_opt with
    | None -> assert false
    | Some in_el ->
	let in_ar =
	  { in_ar = ar;
	    in_next = (fun () -> None);
	    in_postprocess = cp.cp_postprocess;
	    in_close = fun () -> ()
	  } in
	extract_element_as_file
	  ~insecure:true ~prefer_link
	  ~prefix:dir ~preserve_mtime ~overwrite (in_ar,in_el)
	  

let postprocess_copying cp =
  postprocess_ht cp.cp_postprocess
