(* $Id: godi_concmd.ml 953 2011-09-02 21:28:08Z gerd $ *)

(* Format:

   byte pos     bytes                  comment
   0            0x11                   start of sequence
   1            len3 len2 len1 len0    length as decimal digits
   5            c0                     c0 = len0 XOR 0xAA
   6            <message>              ASCII only
   len-1        0x12                   end of sequence

   The <message> is a space-separated list of tokens.

   Command      message
   `Prompt(p,f) "prompt" <space> <base64(p)> <space> <base64(f)>
   `Password_prompt(p,f)
                "password_prompt" <space> <base64(p)> <space> <base64(f)>
 *)

open Printf

type concmd =
    [ `Prompt of string * string
    | `Password_prompt of string * string
    ]

type token =
    [ concmd
    | `Data of string
    | `Incomplete
    ]

let have_console() =
  try Sys.getenv "GODI_CONSOLE_TTY" = "godi"
  with Not_found -> false

let string_of_message m =
  let l = String.length m in
  if l > 9999 - 7 then
    failwith "Godi_concmd.string_of_concmd: too long";
  sprintf "\017%04d%c%s\018"
    (l + 7)
    (Char.chr ((48 + (l+7) mod 10) lxor 0xaa))
    m
    

let string_of_concmd =
  function
    | `Prompt(p,f) ->
	string_of_message
	  (sprintf "prompt %s %s"
	     (Netencoding.Base64.encode p)
	     (Netencoding.Base64.encode f)
	  )
    | `Password_prompt(p,f) ->
	string_of_message
	  (sprintf "password_prompt %s %s"
	     (Netencoding.Base64.encode p)
	     (Netencoding.Base64.encode f)
	  )

exception Incomplete

let digit c =
  match c with
    | '0'..'9' -> ()
    | _ -> raise Not_found

let space_re = Str.regexp " "

let scan b eof =
  if Netbuffer.length b = 0 then
    `Data ""
  else (
    try
      let k = Netbuffer.index_from b 0 '\017' in
      if k = 0 then (
	try
	  if Netbuffer.length b < 7 then raise Incomplete;
	  digit (Netbuffer.get b 1);
	  digit (Netbuffer.get b 2);
	  digit (Netbuffer.get b 3);
	  digit (Netbuffer.get b 4);
	  let len0 = Netbuffer.get b 4 in
	  let c0 = Char.chr((Char.code len0) lxor 0xaa) in
	  if Netbuffer.get b 5 <> c0 then raise Not_found;
	  let l = int_of_string (Netbuffer.sub b 1 4) in
	  if Netbuffer.length b < l then raise Incomplete;
	  if Netbuffer.get b (l-1) <> '\018' then raise Not_found;
	  let m = Netbuffer.sub b 6 (l-7) in
	  let words = Str.split space_re m in
	  ( match words with
	      | [ "prompt"; p64; f64 ] ->
		  let p = Netencoding.Base64.decode p64 in
		  let f = Netencoding.Base64.decode f64 in
		  Netbuffer.delete b 0 l;
		  `Prompt(p,f)
	      | [ "password_prompt"; p64; f64 ] ->
		  let p = Netencoding.Base64.decode p64 in
		  let f = Netencoding.Base64.decode f64 in
		  Netbuffer.delete b 0 l;
		  `Password_prompt(p,f)
	      | _ ->
		  raise Not_found
	  )
	with
	  | Incomplete -> 
	      if eof then raise Not_found;
	      `Incomplete
	  | _ ->
	      let s = Netbuffer.sub b 0 1 in
	      Netbuffer.delete b 0 1;
	      `Data s
      )
      else (
	let s = Netbuffer.sub b 0 k in
	Netbuffer.delete b 0 k;
	`Data s
      )
    with
      | Not_found ->
	  let s = Netbuffer.contents b in
	  Netbuffer.clear b;
	  `Data s
  )


let send_data file data =
  match Sys.os_type with
    | "Win32" ->
	Godi_file.write_file
	  ~mode:`Binary (Godi_file.filename (file ^ ".new")) data;
	Sys.rename (file ^ ".new") file
    | _ ->
	( try Unix.mkfifo file 0o600 
	  with Unix.Unix_error(Unix.EEXIST,_,_) -> ()
	);
	Godi_file.write_file ~mode:`Binary (Godi_file.filename file) data


let receive_data file =
  match Sys.os_type with
    | "Win32" ->
	while not(Sys.file_exists file) do
	  Netsys.sleep 0.1
	done;
	let data =
	  Godi_file.read_file ~mode:`Binary (Godi_file.filename file) in
	Unix.unlink file;
	data
    | _ ->
	( try Unix.mkfifo file 0o600 
	  with Unix.Unix_error(Unix.EEXIST,_,_) -> ()
	);
	Netchannels.with_in_obj_channel
	  (new Netchannels.input_channel (open_in file))
	  (fun ch ->
	     let data = Netchannels.string_of_in_obj_channel ch in
	     Unix.unlink file;
	     data
	  )

